"""Payton Scene Module.

##Main Scene Module:

* SDL2 Window
  * Scene (`payton.scene.scene.Scene`)
    * Geometry (`payton.scene.geometry`)
    * Grid (`payton.scene.grid`)
    * Background (`payton.scene.scene.Background`)
    * Clock (`payton.scene.clock`)
    * Lighting (`payton.scene.light`)
    * Collision Test (`payton.scene.collision`)
    * Camera (`payton.scene.camera`)
    * GUI (`payton.scene.gui`)
    * Material (`payton.scene.material`)
      * Shader (`payton.scene.shader`)
    * Type Support (`payton.scene.types`)

Also, `payton.scene.scene` contents are accessible through `payton.scene` main
module
"""
# pylama:ignore=W
from payton.scene.physics import physics_client
from payton.scene.scene import SHADOW_HIGH, SHADOW_LOW, SHADOW_MID, SHADOW_NONE, Background, Scene

__all__ = ["Scene", "Background", "SHADOW_HIGH", "SHADOW_LOW", "SHADOW_MID", "SHADOW_NONE", "physics_client"]



import math
from itertools import product
from typing import Any

from payton.math.functions import plane_normal
from payton.scene.geometry.mesh import Mesh
from payton.scene.material import DEFAULT

_BULLET = False
try:
    import pybullet

    _BULLET = True
except ModuleNotFoundError:
    _BULLET = False


class Sphere(Mesh):
    def __init__(
        self,
        radius: float = 0.5,
        parallels: int = 12,
        meridians: int = 12,
        **kwargs: Any,
    ) -> None:
        """Initialize a Sphere object with the given parameters

        Keyword arguments:
        radius -- Radius of the sphere
        parallels -- Number of horizontal steps to produce the sphere
        meridians -- Number of vertical steps to produce the sphere
        """
        super().__init__(**kwargs)
        self.radius: float = radius
        self.parallels: int = parallels
        self.meridians: int = meridians
        self.build_sphere()

    def build_sphere(self) -> bool:
        """Build the sphere"""
        r = self.radius
        # step angle is the rotational angle to build the sphere
        step_angle = math.radians(360.0 / self.meridians)
        # step height is the arc in height
        step_height = math.radians(180.0 / self.parallels)
        indices = 0
        u_step = 1.0 / self.meridians
        v_step = 1.0 / self.parallels
        self.clear_triangles()

        for i, j in product(range(self.parallels), range(self.meridians)):
            x1 = r * math.sin(step_height * i) * math.cos(step_angle * j)
            y1 = r * math.sin(step_height * i) * math.sin(step_angle * j)
            z1 = r * math.cos(step_height * i)
            u1 = u_step * j
            v1 = v_step * i

            x2 = r * math.sin(step_height * (i + 1)) * math.cos(step_angle * j)
            y2 = r * math.sin(step_height * (i + 1)) * math.sin(step_angle * j)
            z2 = r * math.cos(step_height * (i + 1))
            u2 = u_step * j
            v2 = v_step * (i + 1)

            x3 = r * math.sin(step_height * (i + 1)) * math.cos(step_angle * (j + 1))
            y3 = r * math.sin(step_height * (i + 1)) * math.sin(step_angle * (j + 1))
            z3 = r * math.cos(step_height * (i + 1))
            u3 = u_step * (j + 1)
            v3 = v_step * (i + 1)

            x4 = r * math.sin(step_height * i) * math.cos(step_angle * (j + 1))
            y4 = r * math.sin(step_height * i) * math.sin(step_angle * (j + 1))
            z4 = r * math.cos(step_height * i)
            u4 = u_step * (j + 1)
            v4 = v_step * i

            normal = plane_normal([x1, y1, z1], [x2, y2, z2], [x3, y3, z3])
            self._vertices.append([x1, y1, z1])
            self._vertices.append([x2, y2, z2])
            self._vertices.append([x3, y3, z3])
            self._vertices.append([x4, y4, z4])
            self._texcoords.append([u1, v1])
            self._texcoords.append([u2, v2])
            self._texcoords.append([u3, v3])
            self._texcoords.append([u4, v4])
            self._normals.append([normal[0], normal[1], normal[2]])
            self._normals.append([normal[0], normal[1], normal[2]])
            self._normals.append([normal[0], normal[1], normal[2]])
            self._normals.append([normal[0], normal[1], normal[2]])
            self._indices.append([indices, indices + 1, indices + 2])
            self._indices.append([indices, indices + 2, indices + 3])
            self.materials[DEFAULT]._indices.append([indices, indices + 1, indices + 2])
            self.materials[DEFAULT]._indices.append([indices, indices + 2, indices + 3])
            indices += 4
        return True

    def _create_collision_shape(self) -> None:
        self._bullet_shape_id = pybullet.createCollisionShape(pybullet.GEOM_SPHERE, radius=self.radius)



from payton.scene.geometry.base import Object
from payton.scene.gui.base import Hud, Text, text_size
from payton.scene.gui.window import Button, Window, WindowAlignment


def help_win() -> Window:
    """Help window."""
    text = """Payton Shortcuts and Mouse Controls:

MOUSE:
Zoom In-Out: Mouse Wheel
Rotate: Right Click and Drag
Pan: Middle Click and Drag
KEYBOARD:
ESC: Quit
H: Show/Hide Help
C: Change Camera mode (Perspective/Orthographic)
Space: Start/Pause Scene Animations/Clocks
G: Show/Hide Grid
W: Display mode (wireframe/solid)
F2: Previous Camera F3: Next Camera
    """
    help_window = Window(
        align=WindowAlignment.RIGHT,
        title="How to use Payton Scene",
        height=10000,
        width=520,
    )
    help_window.add_child("text", Button(width=500, height=400, left=10, top=40, label=text))
    return help_window


def info_box(left: int, top: int, label: str) -> Hud:
    """Create Info Box along with it's HUD so you do not need to setup the
    whole HUD.

    Keyword arguments:
    left -- Left position of the info box
    top -- Top position of the info box
    label -- Label / text of the info box
    """
    hud = Hud()

    res = text_size(label, hud.font)
    width = res[0] + 60
    height = res[1] + 20

    window = Window("Info", width=width, height=height + 120, left=left, top=top)

    window.add_child(
        "label",
        Button(label=label, left=2, width=width, height=height, top=40),
    )

    window.add_child(
        "close",
        Button(
            label="CLOSE",
            left=int(width / 2) - 10,
            top=height + 60,
            width=int(width / 2),
            height=30,
            on_click=window.hide,
        ),
    )
    hud.add_child("window", window)
    return hud


def object_box(left: int, top: int, ob: Object) -> Hud:
    """Show object box."""
    px = ob.matrix[3][0]
    py = ob.matrix[3][1]
    pz = ob.matrix[3][2]
    label = f"""Object:
Position: [{px}, {py}, {pz}]
Number of Indices: {len(ob._indices)}
Number of vertices: {ob._vertex_count}
    """
    hud = Hud()
    width = 400
    height = 200
    window = Window(ob.name, width=width, height=height, left=left, top=top)
    window.add_child("label", Text(position=[10, 20], size=[380, 300], label=label, color=[1, 1, 1]))
    hud.add_child("window", window)
    return hud



"""
ID Software Quake 2 Model File.

Some parts of this file is based on the original work from:
https://github.com/adamlwgriffiths/PyMesh/tree/master/pymesh/md2
"""
import logging
import os
import struct
import time
from copy import deepcopy
from typing import Any, BinaryIO, Dict, Generator, List, NamedTuple, Optional, Tuple, cast

import numpy as np

from payton.scene.geometry.mesh import Mesh
from payton.scene.material import POINTS
from payton.scene.shader import DEFAULT_SHADER, PARTICLE_SHADER, Shader

_SIGNATURE = "IDP2"
_VERSION = 8


class MeshException(Exception):
    """Mesh relatex exceptions for MD2."""

    pass


def _interpolate(mesh_1: Mesh, mesh_2: Mesh, steps: int = 1) -> List[Mesh]:
    """Interpolate two alike meshes.

    This is suitable to fill the blank frames of an animated object
    This function makes the assumption that same indices will be forming
    the same triangle.

    This function returns at least 3 meshes;
    [mesh_1, interpolated mesh, mesh_2]
    """
    if len(mesh_1._vertices) != len(mesh_2._vertices):
        raise MeshException("Mesh 1 and Mesh 2 vertex counts do not match")
    if len(mesh_1.children) > 0:
        raise MeshException("Mesh 1 has children")
    if len(mesh_2.children) > 0:
        raise MeshException("Mesh 2 has children")

    # Generate meshes
    results: List[Mesh] = [mesh_1]

    # Calculate vertex distances;
    distances = []
    for v_i, vertex in enumerate(mesh_1._vertices):
        dist = [0.0, 0.0, 0.0]
        dist[0] = (mesh_2._vertices[v_i][0] - vertex[0]) / (steps + 1)
        dist[1] = (mesh_2._vertices[v_i][1] - vertex[1]) / (steps + 1)
        dist[2] = (mesh_2._vertices[v_i][2] - vertex[2]) / (steps + 1)
        distances.append(dist)

    for i in range(1, steps + 1):
        mesh = Mesh()
        mesh.materials = deepcopy(results[i - 1].materials)
        mesh._texcoords = deepcopy(results[i - 1]._texcoords)
        mesh._indices = deepcopy(results[i - 1]._indices)
        mesh.destroy()  # Destroy references to previous objects opengl
        for v_i, vertex in enumerate(results[i - 1]._vertices):
            mesh._vertices.append(
                [vertex[0] + distances[v_i][0], vertex[1] + distances[v_i][1], vertex[2] + distances[v_i][2]]
            )
        mesh.fix_normals()
        results.append(mesh)

    results.append(mesh_2)

    return results


class MD2Header(NamedTuple):
    """MD2 Model Header Byte Order."""

    ident: bytes = b""
    version: int = 0
    skin_width: int = 0
    skin_height: int = 0
    frame_size: int = 0
    num_skins: int = 0
    num_vertices: int = 0
    num_st: int = 0
    num_tris: int = 0
    num_glcmds: int = 0
    num_frames: int = 0
    offset_skins: int = 0
    offset_st: int = 0
    offset_tris: int = 0
    offset_frames: int = 0
    offset_glcmds: int = 0
    offset_end: int = 0


class MD2Frame(NamedTuple):
    """MD2 Model Frame Information."""

    name: str = ""
    vertices: np.ndarray = np.array([], dtype=np.float64)
    normals: np.ndarray = np.array([], dtype=np.float64)


class MD2TriangleLayout(NamedTuple):
    """MD2 Triangle Collection."""

    vertex_indices: np.ndarray = np.array([], dtype=np.float64)
    tc_indices: np.ndarray = np.array([], dtype=np.float64)


def _read_block(b: BinaryIO, format_str: str, count: int) -> List[Tuple]:
    def chunks(data: bytes, size: int) -> Generator:
        offset = 0
        while offset < len(data):
            yield_size = offset + size
            yield data[offset:yield_size]
            offset += size

    struct_length = struct.calcsize(format_str)
    total_length = struct_length * count
    data = b.read(total_length)
    if len(data) < total_length:
        raise ValueError("MD2: Failed to read '%d' bytes" % (total_length))
    return [struct.unpack(format_str, chunk) for chunk in chunks(data, struct_length)]


def _fix_skin_name(name: bytes) -> str:
    name_str = name.decode("utf-8").replace("\x00", "")
    name_parts = os.path.splitext(name_str)
    if name_parts[1].startswith(".pcx"):
        return f"{name_parts[0]}.pcx"
    return name_str


class MD2(Mesh):
    """
    MD2 File Format Loader.

    I know that this is a pretty old file format. We can consider it as ancient
    On the other hand, our target on creating Payton is not to make a game
    engine.

    Also note that there might be some bugs, I give no guarantee here.
    As this is a generic class, I haven't defined any animations before hand.
    So to make things properly looping in your code, you need to set the
    animation frame ranges yourself. For instance, as you will see at the
    defined example, infantry has a perfect walking loop between frames 2-13

    Unlike original frame names, frames in this class starts from 0 and goes
    incremented by one.

    NOTE: There are some hard-coded pre-assumptions like the scale of the
    objects and they are just defined to look okay on a default scene.

    Example use case:

        .. include:: ../../../examples/basics/26_quake2.py
    """

    def __init__(self, filename: str = "", texture_filename: str = "", **kwargs: Any):
        """Initialize the MD2 Object.

        Keyword arguments:
        filaname -- Filename to load
        texture_filename -- Texture filaname for the MD2 Object
        """
        super().__init__(**kwargs)
        self.header: MD2Header = MD2Header()
        self.triangle_layout = MD2TriangleLayout()
        self.skins: List[str] = []
        self.texture_filename = texture_filename
        self.frames: List[MD2Frame] = []
        self._frame_children: Dict[str, Mesh] = {}
        self.animation: str = ""
        self.animations: Dict[str, List] = {}

        self._active_frame: int = 0
        self._frame_rate: float = 0
        self._from_frame: int = 0
        self._to_frame: int = 0
        self._time: float = 0
        self._path: str = ""
        self._loop: bool = False

        if os.path.exists(filename):
            self.load_file(filename)

    def add_frame_child(self, name: str, mesh: Mesh) -> None:
        """Add Frame Model."""
        self._frame_children[name] = mesh

    def bake_animation(
        self,
        animation_name: str,
        from_frame: int,
        to_frame: int,
        steps: int = 1,
    ) -> None:
        """Original MD2 Format does not include all frames in an animation.

        Instead, there are "key frames". There are gaps between the key-frames
        so you need to fill-in the empty frames by calculating the motion
        between key-frames. Bake Animation method simply fills these gaps by
        calculating the N steps between key-frames. This gives you a smooth
        animation.

        Keyword arguments:
        animation_name -- Animation name to bake
        from_frame -- Starting frame for baking
        to_frame -- Ending frame for baking
        steps -- Number of steps to interpolate.
        """
        if animation_name not in self.animations:
            logging.error(f"Animation {animation_name} not found in object")
            return
        anim = self.animations[animation_name]
        if from_frame < anim[0] or from_frame > anim[1]:
            logging.error("from_frame out of bounds")
            return
        if to_frame < anim[0] or to_frame > anim[1]:
            logging.error("to_frame out of bounds")
            return

        frames = []
        for i in range(from_frame, to_frame):
            frame_name = f"{animation_name}{i}"
            frame = self._frame_children[frame_name]
            nf = i + 1
            next_name = f"{animation_name}{nf}"
            next_frame = self._frame_children[next_name]

            interpolated_frames = _interpolate(frame, next_frame, steps)
            frames.extend(interpolated_frames[0:-1])
        last_frame = self._frame_children[f"{animation_name}{to_frame}"]
        first_frame = self._frame_children[f"{animation_name}{from_frame}"]
        join_frames = _interpolate(last_frame, first_frame, steps)
        frames.extend(join_frames)

        for i in range(len(frames)):
            frame_name = f"{animation_name}{i}"
            self._frame_children[frame_name] = frames[i]
        self.animations[animation_name] = [0, len(frames) - 1]

    def animate(
        self,
        animation_name: str,
        from_frame: int,
        to_frame: int,
        loop: bool = True,
    ) -> None:
        """Set the model in motion.

        Keyword arguments
        animation_name -- Name of the animation to play
        from_frame -- Loop starting frame
        to_frame -- Loop ending frame
        loop -- Loop the animation?
        """
        self.animation = ""  # For thread safety
        if animation_name not in self.animations:
            logging.error(f"Animation {animation_name} not found in object")
            return
        anim = self.animations[animation_name]
        if from_frame < anim[0] or from_frame > anim[1]:
            logging.error("from_frame out of bounds")
            return
        if to_frame < anim[0] or to_frame > anim[1]:
            logging.error("to_frame out of bounds")
            return
        self._loop = loop
        self._active_frame = from_frame
        self._from_frame = from_frame
        self._to_frame = to_frame
        self.animation = animation_name
        self._frame_rate = 0
        for child in self.children:
            child_obj = cast(MD2, self.children[child])
            child_obj.animate(animation_name, from_frame, to_frame, loop)

    @property
    def frame(self) -> str:
        """Get the current frame name as <animation_name><frame_number>."""
        if self._time == 0:
            self._time = time.time()

        tdiff = time.time() - self._time
        if self._frame_rate == 0:
            self._frame_rate = 1 / (self._to_frame - self._from_frame)

        if tdiff >= self._frame_rate:
            self._active_frame += 1
            if self._active_frame > self._to_frame:
                if self._loop:
                    self._active_frame = self._from_frame
                else:
                    self._active_frame -= 1

            self._time = time.time()

        return f"{self.animation}{self._active_frame}"

    def render(
        self,
        lit: bool,
        shader: Shader,
        parent_matrix: Optional[np.ndarray] = None,
        _primitive: Optional[int] = None,
    ) -> None:
        """Render cycle for the MD2 Object.

        Keyword arguments:
        lit -- Is this an illuminated object?
        shader -- Shader to use for rendering
        parent_matrix -- Matrix of the parent object if exists
        _primitive -- Enforced OpenGL primitive to render
        """
        if not self._visible:
            return

        self.update_matrix(parent_matrix=parent_matrix)
        self.track()

        # Render motion path
        if self.track_motion:
            self._motion_path_line.render(lit, shader, parent_matrix)

        if self.animation == "":
            next(iter(self._frame_children.values())).render(lit, shader, self._model_matrix)
            for child in self.children:
                self.children[child].render(lit, shader, self._model_matrix)
            return

        self._frame_children[self.frame].render(lit, shader, self._model_matrix)
        for child in self.children:
            self.children[child].render(lit, shader, self._model_matrix)

    def load_file(self, filename: str) -> None:
        """Load MD2 file.

        Keyword arguments:
        filaname -- MD2 Filaname to load.
        """
        if not os.path.exists(filename):
            raise BaseException(f"File not found: {filename}")
        self._path = os.path.dirname(os.path.abspath(filename))
        with open(filename, "rb") as f:
            self.load_buffer(f)

    def load_buffer(self, f: BinaryIO) -> None:
        """Load MD2 buffer into memory and compile the object."""
        self.read_header(f)
        self.read_skin(f)
        self.read_tex_coords(f)
        self.read_triangles(f)
        self.load_frames(f)
        self.compile()

    def read_triangles(self, f: BinaryIO) -> None:
        f.seek(self.header.offset_tris, os.SEEK_SET)
        triangles = np.array(_read_block(f, "< 6H", self.header.num_tris), dtype=np.uint16)
        triangles.shape = (-1, 6)
        vertex_indices = triangles[:, :3]
        tc_indices = triangles[:, 3:]
        self.triangle_layout = MD2TriangleLayout(vertex_indices=vertex_indices, tc_indices=tc_indices)

    def compile(self) -> None:
        self.animations = {}
        for frame in self.frames:
            name = "".join(i for i in frame.name if not i.isdigit())
            if name not in self.animations:
                self.animations[name] = [0, -1]
            self.animations[name][1] += 1
            num = self.animations[name][1]
            frame_name = f"{name}{num}"
            self.build_frame(frame, frame_name)

    def build_frame(self, frame_information: MD2Frame, name: str) -> None:
        mesh = Mesh()
        for i, tri in enumerate(self.triangle_layout.vertex_indices):
            v3 = frame_information.vertices[tri[1]].tolist()
            v2 = frame_information.vertices[tri[0]].tolist()
            v1 = frame_information.vertices[tri[2]].tolist()
            v1[0], v1[1] = v1[1], v1[0]
            v2[0], v2[1] = v2[1], v2[0]
            v3[0], v3[1] = v3[1], v3[0]

            t3 = self._texcoords[self.triangle_layout.tc_indices[i][1]]
            t2 = self._texcoords[self.triangle_layout.tc_indices[i][0]]
            t1 = self._texcoords[self.triangle_layout.tc_indices[i][2]]

            mesh.add_triangle(
                vertices=[v1, v2, v3],
                texcoords=[t1, t2, t3],
            )

        mesh.fix_normals(False)
        if self.texture_filename != "":
            mesh.material.texture = os.path.join(self._path, os.path.basename(self.texture_filename))
            mesh.material.particle_size = 0.1
        elif len(self.skins) > 0:
            mesh.material.texture = os.path.join(self._path, os.path.basename(self.skins[0]))
            mesh.material.particle_size = 0.1

        self.add_frame_child(name, mesh)

    def read_header(self, f: BinaryIO) -> None:
        self.header = MD2Header._make(_read_block(f, "< 4s16l", 1)[0])

        if self.header.ident.decode("ascii") != _SIGNATURE:
            raise BaseException("MD2 Identifier is incorrect")
        if self.header.version != _VERSION:
            raise BaseException("Invalid Version")

    def read_skin(self, f: BinaryIO) -> None:
        f.seek(self.header.offset_skins, os.SEEK_SET)
        skin_struct = struct.Struct("< %s" % ("64s" * self.header.num_skins))

        self.skins = [_fix_skin_name(skin) for skin in skin_struct.unpack(f.read(skin_struct.size))]

    def read_tex_coords(self, f: BinaryIO) -> None:
        f.seek(self.header.offset_st, os.SEEK_SET)
        tcs = np.array(_read_block(f, "< 2h", self.header.num_st), dtype=np.float64)
        tcs.shape = (-1, 2)
        tcs /= [float(self.header.skin_width), float(self.header.skin_height)]
        self._texcoords = tcs.tolist()

    def load_frames(self, f: BinaryIO) -> None:
        f.seek(self.header.offset_frames, os.SEEK_SET)
        self.frames = [self.read_frame(f) for x in range(self.header.num_frames)]

    def read_frame(self, f: BinaryIO) -> MD2Frame:
        frame_translations = np.array(_read_block(f, "< 3f", 2), dtype=np.float64)
        scale = frame_translations[0]
        translation = frame_translations[1]

        name = _read_block(f, "< 16s", 1)[0][0]
        name = str(name).split("\x00")[0].split("\\")[0].replace("b'", "")

        frame_vertex_data = np.array(_read_block(f, "<4B", self.header.num_vertices), dtype=np.uint8)

        frame_vertex_data.shape = (-1, 4)

        vertices_short = frame_vertex_data[:, :3]
        vertices = vertices_short.astype(np.float64)
        vertices.shape = (-1, 3)

        vertices *= scale
        vertices += (translation[0], translation[1], 0)
        vertices /= 20.0

        return MD2Frame(name=name, vertices=vertices)

    def to_dict(self) -> Dict[str, Any]:
        """Convert the MD2 Object to Dictionary"""
        animation = self.animation
        if animation == "":
            animation = list(self.animations.keys())[0]

        frame_name = f"{animation}{self._active_frame}"
        result = self._frame_children[frame_name].to_dict()
        result["matrix"] = self.matrix
        result["children"] = {name: self.children[name].to_dict() for name in self.children}
        return result

    def set_texture(self, texture_filename: str) -> None:
        """Set the object texture filaname"""
        for m in self.children.values():
            m.material.texture = texture_filename
            m.material.refresh()

    def toggle_wireframe(self) -> None:
        """Toggle wireframe view of the Object"""
        d = (self.material.display + 1) % 3

        for mat in self.materials.values():
            mat.display = d
            mat.particle_size = 0.01

        self.shader = PARTICLE_SHADER if d == POINTS else DEFAULT_SHADER
        self.refresh()
        for frame in self._frame_children.values():
            frame.toggle_wireframe()



"""Camera module
"""

import logging
import math
from typing import Any, Dict, Optional, Tuple

import numpy as np
import pyrr

from payton.math.functions import add_vectors, sub_vector, vector_transform_4
from payton.math.geometry import raycast_plane_intersect
from payton.math.vector import Vector3D
from payton.scene.geometry.base import Object

BUTTON_LEFT = 1
BUTTON_RIGHT = 2
BUTTON_MIDDLE = 3


class Camera:
    def __init__(
        self,
        position: Optional[Vector3D] = None,
        target: Optional[Vector3D] = None,
        up: Optional[Vector3D] = None,
        target_object: Optional[Object] = None,
        fov: float = 45.0,
        aspect_ratio: float = 1.33333,
        near: float = 0.1,
        far: float = 100.0,
        zoom: float = 10.0,
        active: bool = False,
        perspective: bool = True,
        viewport_size: Optional[Vector3D] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the camera

        Keyword arguments:
        position -- Position of the camera in space
        target -- Where does the camera point?
        up -- Up direction of the camera (roll)
        target_object -- Camera will follow the target object if defined
        fov -- Field of view in degrees
        aspect_ratio -- Aspect ratio of the camera
        near -- Nearest distance that the camera can see
        far -- Furthest distance that the camera can see (effects performance)
        zoom -- Zoom ratio
        active -- Is this the active camera in the scene?
        perspective -- Perspective camera if True, otherwise Orthographic camera (default True)
        """
        self.position = [10.0, 10.0, 5.0] if position is None else position
        self.target = [0.0, 0.0, 0.0] if target is None else target
        self.up = [0.0, 0.0, 1.0] if up is None else up

        self.target_object: Optional[Object] = target_object
        self._previous_target_location: Optional[Vector3D] = None
        self.fov: float = fov
        self.aspect_ratio: float = aspect_ratio
        self._near: float = near
        self._far: float = far
        self.min_elevation = -1
        self.max_elevation = -1
        self.min_distance = 0
        self.max_distance = -1

        # zoom factor for Orthographic projection
        self._zoom: float = zoom
        self.active: bool = active
        self._perspective: bool = perspective
        self._use_cache = False
        self._viewport_size = [0.0, 0.0]
        if viewport_size:
            self._viewport_size = viewport_size

        # Store matrices for future reference.
        self._projection: Optional[np.ndarray] = None
        self._view: Optional[np.ndarray] = None
        self._prev_intersection: Optional[np.ndarray] = None

    def to_dict(self) -> Dict[str, Any]:
        """Convert the camera to a dictionary"""
        return {
            "position": self.position,
            "target": self.target,
            "up": self.up,
            "fov": self.fov,
            "aspect_ratio": self.aspect_ratio,
            "zoom": self._zoom,
            "near": self._near,
            "far": self._far,
            "perspective": self.perspective,
        }

    def distance(self) -> float:
        """Calculate distance to target and return as a float"""
        xdiff = self.position[0] - self.target[0]
        ydiff = self.position[1] - self.target[1]
        zdiff = self.position[2] - self.target[2]
        res = math.sqrt(xdiff * xdiff + ydiff * ydiff + zdiff * zdiff)
        if res == 0:
            res = 0.001
        return res

    @property
    def perspective(self) -> bool:
        """Is this a perspective projected camera"""
        return self._perspective

    @perspective.setter
    def perspective(self, perspective: bool) -> None:
        """Set perspective

        Keyword arguments:
        perspective -- Boolean (True for perspective, False for orthographic)
        """
        self._perspective = perspective
        self._use_cache = False

    @property
    def zoom(self) -> float:
        """Return the zoom factor as a float"""
        return self._zoom

    @zoom.setter
    def zoom(self, zoom_ratio: float) -> None:
        """Set the zoom factor

        Keyword arguments:
        zoom_ratio -- Zoom ratio to be set
        """
        self._zoom = zoom_ratio
        self._use_cache = False

    @property
    def near(self) -> float:
        """Return the near plane distance"""
        return self._near

    @near.setter
    def near(self, distance: float) -> None:
        """Set the near plane distance

        Keyword arguments:
        distance -- Distance in units
        """
        self._near = distance
        self._use_cache = False

    @property
    def far(self) -> float:
        """Return the far plane distance"""
        return self._far

    @far.setter
    def far(self, distance: float) -> None:
        """Set the far plane distance

        Keyword arguments:
        distance -- Distance in units
        """
        self._far = distance
        self._use_cache = False

    def rotate_around_target(self, phi: float, theta: float) -> None:
        """Rotate the camera around the target.

        NOTE: This method works with Spherical Coordinates!
        https://en.wikipedia.org/wiki/Spherical_coordinate_system
        Keyword arguments:
        phi -- Phi angle in radians
        theta -- Theta angle in radians"""

        diff = sub_vector(self.position, self.target)

        r = self.distance()
        current_theta = math.acos(diff[2] / r)
        current_phi = math.atan2(diff[1], diff[0])

        if self.min_elevation != -1 and current_theta < self.min_elevation:
            current_theta = self.min_elevation

        if self.max_elevation != -1 and current_theta > self.max_elevation:
            current_theta = self.max_elevation

        ct = current_theta + math.radians(theta)
        cp = current_phi + math.radians(phi)
        ct = max(ct, 0.001)
        ct = min(ct, 3.13)

        x = r * math.sin(ct) * math.cos(cp)
        y = r * math.sin(ct) * math.sin(cp)
        z = r * math.cos(ct)

        self.position = [
            x + self.target[0],
            y + self.target[1],
            z + self.target[2],
        ]
        self._use_cache = False

    def pan(self, x: int, y: int, w: int, h: int) -> None:
        """Pan the camera to sideways

        Keyword arguments:
        x -- X displacement
        y -- Y displacement
        w -- Width of the viewport
        h -- Height of the viewport
        """
        if self.target_object is not None:
            logging.error("Camera has an active target, can not pan")
            return
        px = math.ceil(w / 2)
        py = math.ceil(h / 2)
        _ceye, center_vector = self.screen_to_world(px, py, w, h)
        eye, vector = self.screen_to_world(x, y, w, h)
        starget = self.target
        hit = raycast_plane_intersect(eye, vector, starget, center_vector)
        if hit is None:
            return
        if self._prev_intersection is None:
            self._prev_intersection = hit
        else:
            diff = np.subtract(self._prev_intersection, hit)
            self.position = [
                self.position[0] + diff[0],
                self.position[1] + diff[1],
                self.position[2] + diff[2],
            ]
            self.target = [
                self.target[0] + diff[0],
                self.target[1] + diff[1],
                self.target[2] + diff[2],
            ]
        self._use_cache = False

    def mouse_move(
        self,
        button: int,
        shift: bool,
        ctrl: bool,
        x: int,
        y: int,
        xrel: int,
        yrel: int,
        w: int,
        h: int,
    ) -> None:
        """Mouse move event handler for rotation and panning

        Keyword arguments:
        button -- Pressed button
        shift -- If the shift key is down
        ctrl -- If the ctrl key is down
        x -- x coordinate of the cursor
        y -- y coordinate of the cursor
        xrel -- x displacement of the cursor
        yrel -- y displacement of the cursor
        w -- Viewport width
        h -- Viewport height
        """
        if button == BUTTON_RIGHT:
            self.rotate_around_target(-xrel, -yrel)

        if button == BUTTON_MIDDLE:
            self.pan(x, y, w, h)

    def mouse_wheel(self, yrel: int) -> None:
        """Mouse wheel event handler for zoom

        Keyword arguments:
        yrel -- y displacement of the cursor
        """
        if self.perspective:
            self.distance_to_target(self.distance() + yrel)
        else:
            self.distance_to_target(self.zoom + yrel)

    def distance_to_target(self, distance: float) -> None:
        """Set the distance to the target

        NOTE: This refers to the radius in spherical coordinates

        Keyword arguments:
        distance -- Distance to target
        """
        if not self.perspective:
            self.zoom = distance
            return

        if self.max_distance != -1:
            if distance > self.max_distance:
                distance = self.max_distance
        if distance < self.min_distance:
            distance = self.min_distance

        diff = sub_vector(self.position, self.target)
        _theta = math.acos(diff[2] / self.distance())
        _phi = math.atan2(diff[1], diff[0])

        self.position = [
            self.target[0] + distance * math.sin(_theta) * math.cos(_phi),
            self.target[1] + distance * math.sin(_theta) * math.sin(_phi),
            self.target[2] + distance * math.cos(_theta),
        ]
        self._use_cache = False

    def render(self) -> Tuple[np.ndarray, np.ndarray]:
        """Render the camera."""
        if self._use_cache and self.target_object is None and self._projection is not None and self._view is not None:
            return self._projection, self._view

        if self.perspective:
            proj_matrix = pyrr.matrix44.create_perspective_projection(
                self.fov,
                self.aspect_ratio,
                self.near,
                self.far,
                dtype=np.float32,
            )
        else:
            x = 70 * (self.aspect_ratio)
            y = 70
            if self.zoom == 0:
                self.zoom = 0.01
            self.near = max(self.near, 1)
            proj_matrix = pyrr.matrix44.create_orthogonal_projection_matrix(
                left=(-x / self.zoom),
                right=(x / self.zoom),
                top=(y / self.zoom),
                bottom=(-y / self.zoom),
                near=self.near,
                far=self.far,
                dtype=np.float32,
            )

        self._projection = proj_matrix
        eye = np.array(list(self.position), dtype=np.float32)

        if self.target_object:
            # I believe there is a bug at mypy about @property methods
            if self._previous_target_location is None:
                self._previous_target_location = self.target_object.position
            target_diff = sub_vector(self.target_object.position, self._previous_target_location)
            self.position = add_vectors(self.position, target_diff)
            eye = np.array(list(self.position), dtype=np.float32)
            self.target = self.target_object.position
            self._previous_target_location = self.target_object.position

        target = np.array(list(self.target), dtype=np.float32)
        up = np.array(list(self.up), dtype=np.float32)
        view_matrix = pyrr.matrix44.create_look_at(eye[:3], target[:3], up[:3])
        self._view = view_matrix
        self._use_cache = True
        return proj_matrix, view_matrix

    def world_to_screen(self, world_coordinates: Vector3D) -> Vector3D:
        """Turn the world coordinates into screen coordinates.

        Keyword arguments:
        world_coordinates -- 3D Vector in space coordinates

        Returns:
        3D vector coordinates.
        """
        if len(world_coordinates) == 3:
            world_coordinates.append(1)
        proj_matrix, view_matrix = self.render()
        clip_space = vector_transform_4(world_coordinates, view_matrix)
        clip_space = vector_transform_4(clip_space, proj_matrix)
        div = clip_space[3]
        clip_space = [elem / div for elem in clip_space]
        clip_space = [
            self._viewport_size[0] * (clip_space[0] + 1) / 2.0,
            self._viewport_size[1] - (self._viewport_size[1] * (clip_space[1] + 1) / 2.0),
            0,
        ]
        return clip_space

    def screen_to_world(self, x: int, y: int, width: int, height: int) -> Tuple[np.ndarray, np.ndarray]:
        """Turn the screen coordinates into world coordinates.

        Imagine a point on the surface of the camera (your cursor),
        this method turns X-Y coordinates into world coordinates

        Keyword arguments:
        x -- x coordinate of the cursor
        y -- y coordinate of the cursor
        width -- Width of the viewport
        height -- Height of the viewport
        """
        x_f = (2.0 * x) / width - 1.0
        y_f = 1.0 - (2.0 * y) / height

        eye = np.array(
            [self.position[0], self.position[1], self.position[2], 1.0],
            dtype=np.float32,
        )

        ray_start = np.array([x_f, y_f, 1.0, 1.0], dtype=np.float32)
        proj = self._projection
        inv_proj = pyrr.matrix44.inverse(proj)
        eye_coords_ray = pyrr.matrix44.apply_to_vector(inv_proj, ray_start)
        eye_coords = np.array([eye_coords_ray[0], eye_coords_ray[1], -1.0, 0.0], dtype=np.float32)
        view = self._view
        inv_view = pyrr.matrix44.inverse(view)
        ray_end = pyrr.matrix44.apply_to_vector(inv_view, eye_coords)

        if not self.perspective:
            ray_start = np.array([x_f, y_f, 1.0, 1.0], dtype=np.float32)
            eye_coords_ray = pyrr.matrix44.apply_to_vector(inv_proj, ray_start)
            eye_coords = np.array(
                [eye_coords_ray[0], eye_coords_ray[1], 0.0, 0.0],
                dtype=np.float32,
            )
            eye = pyrr.matrix44.apply_to_vector(inv_view, eye_coords)
            ray = ray_end - eye
            ray_dir = pyrr.vector.normalize(ray[0:4])
            eye[0] -= ray_dir[0] * 500
            eye[1] -= ray_dir[1] * 500
            eye[2] -= ray_dir[2] * 500
            eye[3] = 1

            return eye, ray_dir

        ray_dir = pyrr.vector.normalize(ray_end[0:4])
        return (eye, ray_dir)
